/********************************************************
 * cd_io_b -- Backup things to a CD -- this does the 	*
 * 	I/O						*
 *********************************************************/
#include <sys/param.h>
#include <cstdio>
#include <unistd.h>
#include <glob.h>
#include <mntent.h>

#include "backup.h"
#include "util.h"
#include "backup_info.h"
#include "tmp_file.h"
#include "log.h"
#include "proc.h"
#include "tty_lock.h"
#include "cd_io_b.h"

static unsigned int next_cd_number = 1;	// Current CD number

// Map CD file to batch file
typedef std::map<std::string, std::string> cd_map;

static cd_map batch_to_cd;	// Key = cd name, value = batch name

/********************************************************
 * cd_label << -- Write out a label object		*
 ********************************************************/
std::ostream &operator << (
    std::ofstream& label_file, 
    const cd_label &the_label
)
{
    label_file << CD_LABEL << '\n';
    label_file << CD_LABEL_SERIAL << ": " << the_label.serial << '\n';
    label_file << CD_LABEL_DATE   << ": " << the_label.date << '\n';
    label_file << CD_LABEL_CD     << ": " << the_label.cd_number << '\n';
    label_file << CD_LABEL_LAST   << ": " << the_label.last << '\n';
    return (label_file);
}
/********************************************************
 * cd_io_b -- Create a new CD I/O object		*
 * 							*
 * This object takes batches of files, creates an ISO	*
 * image and writes the result a cd.			*
 ********************************************************/
cd_io_b::cd_io_b(
    const std::string& name,	// The name of this IO
    const config& config	// Global configuration
):
	cd_io_c(name, config), first(true)
{
    image_command = config.get_option(SECTION_IO, name, IO_CD_IMAGE);
    burn_command  = config.get_option(SECTION_IO, name, IO_CD_BURN);
    cd_size = convert_size(config.get_option(SECTION_IO, name, IO_CD_SIZE));
    if (config.defined(SECTION_IO, name, IO_CD_PRE_IMAGE)) {
	pre_image = config.get_option(SECTION_IO, name, IO_CD_PRE_IMAGE);
    } else {
	pre_image = "";
    }

    // Create meta data for this CD only
    // (Destroyed when CD is burned)
    cd_meta = NULL;
    new_cd();
}
/********************************************************
 * cd_io_b::write -- Write out a batch file.  When 	*
 * 	enough files are obtained, an image is burned.	*
 ********************************************************/
void cd_io_b::write(
    const batch_info_b* const batch_info	// Information about the batch
) 
{
    file_list.insert(batch_info->batch_file);

    struct stat file_stat;	// The stat for the file

    if (stat(batch_info->batch_file.c_str(), &file_stat) != 0) {
	INTERNAL_ERROR(std::string("Batch file: ") + 
		batch_info->batch_file + " disappeared");
    }
    // The name of the file as it appears on cd
    char cd_file[MAXPATHLEN];

    sprintf(cd_file, "data_%04d.dat", file_sequence);

    batch_to_cd[batch_info->batch_file] = std::string(cd_file);

    // The information about this CD
    cd_io_info_b* io_info = 
	new cd_io_info_b(cd_number, file_sequence, cd_file);

    fill_in_head(batch_info->batch_file, io_info);

    global_meta.add_batch(batch_info, io_info);
    cd_meta->add_batch(batch_info, io_info);

    ++file_sequence;
    current_size += file_stat.st_size;
    if (current_size < cd_size) {
	return;
    }
    finish_cd("", "");
    current_size = 0;
}
/********************************************************
 * write_label -- Write out the CD label file		*
 ********************************************************/
void cd_io_b::write_label(
    const bool last	// True if this is the last CD
) {
    // The file the label is written into
    std::ofstream label_file((image_dir + "/label.txt").c_str());

    if (label_file.bad()) {
	SYSTEM_ERROR("Could not open label file for CD");
    }
    cd_label the_label(cd_number, last);
    label_file << the_label;
    
    label_file.close();
}
/********************************************************
 * finish_cd -- Write out a CD				*
 ********************************************************/
void cd_io_b::finish_cd(
    const std::string global_meta_file,	// File to write global meta data
    const std::string meta_report	// Report file for meta data
) {
    global_meta.save(global_meta_file, META_FULL);
    global_meta.report(meta_report);

    char image_dir_base[MAXPATHLEN];	// Image directory basename
    sprintf(image_dir_base, "image.%d", cd_number);

    // Get directory for the image
    image_dir = tmp_name(image_dir_base);

    if (mkdir(image_dir.c_str(), 0700) != 0) {
	SYSTEM_ERROR("Could not create image directory: " + image_dir);
    }

    write_label(global_meta_file != "");
    
    // Loop through each batch file
    for (string_set::const_iterator batch = file_list.begin();
	    batch != file_list.end();
	    ++batch) {

	// Turn the file name into a cd name
	cd_map::const_iterator cd_name = batch_to_cd.find(*batch);
	if (cd_name == batch_to_cd.end()) {
	    INTERNAL_ERROR("No CD name for " + *batch);
	}

	// Name (full path) of the file on cd
	std::string full_cd_name = (image_dir + "/" + cd_name->second);
	if (symlink(batch->c_str(), full_cd_name.c_str()) != 0) {
	    SYSTEM_ERROR("Could not create symlink " +
		*batch + " -> " + full_cd_name);
	}
    }

    if (global_meta_file != "") {
	// Name (full path) of the file on cd
	std::string full_cd_name = (image_dir + "/meta.global");

	if (symlink(global_meta_file.c_str(), full_cd_name.c_str()) != 0) {
	    SYSTEM_ERROR("Could not create symlink " +
		global_meta_file + " -> " + full_cd_name);
	}

	// Link in the global report
	full_cd_name = (image_dir + "/global_report.txt");

	if (symlink(meta_report.c_str(), full_cd_name.c_str()) != 0) {
	    SYSTEM_ERROR("Could not create symlink " +
		meta_report + " -> " + full_cd_name);
	}
    }
    cd_meta->save(image_dir + "/meta.cd", "CD Metadata");
    cd_meta->report(image_dir + "/cd_report.txt");

    char iso[MAXPATHLEN];	// The base of the iso image file
    sprintf(iso, "iso.%d", cd_number);

    // The name of the image file
    std::string image_file = tmp_name(iso);
    char cmd[MAXPATHLEN + 500];	// The command as a string

    sprintf(cmd, "cd %s;date=`date`;"
	    "cd_number=%d;backup_serial=%ld;"
	    "image_file=%s;%s",
	    image_dir.c_str(), 
	    cd_number, 
	    backup_serial(),
	    image_file.c_str(), 
	    image_command.c_str());

    if (!first) {
	while (1) {
	    {
		tty_lock the_lock;	// Lock the console

		std::cout << "Load cd number " << cd_number << " in " << mount_dev 
		    << " and press <enter>" << std::flush;
		std::string response;
		getline(std::cin, response);
	    }

	    // We expect an error
	    try {
		mount(-1);

		cd_label label;	// The label of the cd in the drive
		label.read_label(mount_dir);
		umount();
		if (label.serial == backup_serial()) {
		    std::cout << "Attempt to reused a cd " << std::endl;
		    continue;
		}
		break;
	    }
	    catch (...) {
		break;
	    }
	}
    }
    first = false;

    if (pre_image != "") {
	do_system("image_dir=" + image_dir + ";" + pre_image, LOG_IO);
    }
    do_system(cmd, LOG_IO);

    if (! keep_option) {
	glob_t glob_info;	// Information about the glob

	if (glob((image_dir + "/*").c_str(), 0, NULL, &glob_info) != 0) {
	    SYSTEM_ERROR("Glob of " + image_dir + "/* failed");
	}

	for (unsigned int i =0; i < glob_info.gl_pathc; ++i) {
	    if (unlink(glob_info.gl_pathv[i]) != 0)
		SYSTEM_ERROR("Could not unlink: " + 
			std::string(glob_info.gl_pathv[i]));
	}
	globfree(&glob_info);
	if (rmdir(image_dir.c_str()) != 0)
	    SYSTEM_ERROR("Could not remove: " + image_dir);

	for (string_set::const_iterator batch = file_list.begin();
		batch != file_list.end();
		++batch) {
	    if (unlink(batch->c_str()) != 0)
		SYSTEM_ERROR("Could not remove: " + *batch);
	}
    }
    umount();

    sprintf(cmd, "image_file=%s;%s", 
	    image_file.c_str(),
	    burn_command.c_str());
    do_system(cmd, LOG_IO);

    if (! image_option) {
	if (unlink(image_file.c_str()) != 0) {
	    SYSTEM_ERROR("Unable to unlink image file " + image_file);
	}
    }
    new_cd();
}

/******************************************************** 
 * new_cd -- Called when we start a new cd		*
 ********************************************************/
void cd_io_b::new_cd(void)
{
    file_sequence = 0;
    current_size = 0;
    file_list.clear();

    if (cd_meta != NULL) {
	delete (cd_meta);
	cd_meta = NULL;
    }
    cd_meta = new meta_b;

    cd_number = next_cd_number;
    ++next_cd_number;
}

/********************************************************
 * finish -- Finish writing to a CD			*
 ********************************************************/
void cd_io_b::finish(
    const std::string meta_file, 	// The file containing the meta data
    const std::string meta_report	// The file containing the meta report
) 
{
    global_meta.save(meta_file, META_FULL);
    global_meta.report(meta_report);
    finish_cd(meta_file, meta_report);
}
