/********************************************************
 * collection_batch_b -- Module to backup files in a 	*
 * collection.						*
 ********************************************************/
#include <dirent.h>
#include <unistd.h>
#include <sys/fcntl.h>

#include "magic.h"
#include "util.h"
#include "tmp_file.h"
#include "backup.h"
#include "log.h"
#include "collect_batch_b.h"
#include "backup_info.h"
#include "crc32.h"
#include "proc.h"

/********************************************************
 * collection_batch_info -- Information on a collection	*
 ********************************************************/
class collection_batch_info_b: public batch_info_b {
    public:
	explicit collection_batch_info_b(
	     const string_set& i_files,
	     const std::string i_batch_file) :
	    batch_info_b(BATCH_TYPE_COLLECTION, 
		    i_files, i_batch_file) {}
	// Destructor defaults
	// Copy constructor defaults
	// Assignment operator defaults
	void write(std::ostream &out_file) const;
};
/********************************************************
 * collection_batch::collection_batch -- 		*
 * 	Create a batching object			*
 ********************************************************/
collection_batch_b::collection_batch_b(
    const std::string& batch_name,	// The name of the batch
    const config& config		// The configuration information
) : batch_b(batch_name, config)
{
    current_size = 0;
    if (names_needed) {
	batch_limit = convert_size(config.get_option(SECTION_BATCH, 
		    batch_name, BATCH_SIZE));

	file_limit =  convert_size(config.get_option(SECTION_BATCH, 
		    batch_name, BATCH_COLLECT_FILE_SIZE));

	if (batch_limit <= file_limit) {
	    USER_ERROR(std::string(BATCH_SIZE) + " must be bigger than " + 
		    BATCH_COLLECT_FILE_SIZE);
	}
    }
}

/********************************************************
 * collection_batch::add_files -- Add a bunch of files 	*
 * to the batch.					*
 ********************************************************/
batch_b::status collection_batch_b::add_files(
    filer &the_filer	// The filer which contains the files to use
)
{
    if (! names_needed) {
	if (done) {
	    return (BATCH_NONE);
	} else {
	    return (BATCH_FULL);
	}
    }
    status batch_status = BATCH_NONE;	// Assume we added no files

    // The files we may want to add
    string_set files_to_add = the_filer.get_files();

    for (string_set::const_iterator cur_file = files_to_add.begin();
	 cur_file != files_to_add.end();
	 ++cur_file) {
	// Did we do this one already
	if (the_filer.done(*cur_file))
	    continue;

	struct stat64 stat;	// Stat info for the file

	if (lstat64(cur_file->c_str(), &stat) != 0) {
	    log_msg(LOG_WARNING, "Warning: File " + 
		    *cur_file + 
		    " disappeared(errno=" + int_to_string(errno) + ")");
	    // If it is no longer here, it can not be backed up
	    the_filer.mark(*cur_file);
	    continue;
	}

	size_t size = stat.st_size;	// The size of the file

	if ((size > file_limit) && (S_ISREG(stat.st_mode))) {
	    // The directory
	    log_msg(LOG_BATCH, "File too big " + *cur_file);
	    continue;
	}

	if (size + current_size > batch_limit) {
	    log_msg(LOG_BATCH, "Returning BATCH_FULL");
	    return (BATCH_FULL);
	}

	current_size += size;

	// The list of files which have this inode
	string_set files = the_filer.all_names(*cur_file);

	for (string_set::const_iterator add_file = files.begin();
	     add_file != files.end();
	     ++add_file)
	{
	    current_files.insert(*add_file);
	    log_msg(LOG_BATCH, "Adding file to batch " + *add_file);
	}
	the_filer.mark(files);

	// We have at least a partial list now
	batch_status = BATCH_PART;
    }
    switch (batch_status) {
	case BATCH_PART: 
	    log_msg(LOG_BATCH, "Returning BATCH_PART");
	    break;
	case BATCH_NONE: 
	    log_msg(LOG_BATCH, "Returning BATCH_NONE");
	    break;
	default:
	    INTERNAL_ERROR("Crazy status");
	    break;
    }
    return (batch_status);
}

/********************************************************
 * collection_batch::create_batch -- 			*
 * 		Create the batch file			*
 * 							*
 * Returns information about the batch.			*
 ********************************************************/
batch_info_b* collection_batch_b::create_batch(void) 
{
    if (names_needed && current_files.empty())
	return (NULL);

    if ((!names_needed) && done)
	return (NULL);

    if (done == false);

    // Name the batch of files
    std::string batch_file = tmp_name("collection.file");

    std::string cmd; // The command to create the collection

    // The name of the file to hold the names
    std::string name_file_name;

    if (names_needed) {
	// The name of the file to hold the names
	name_file_name = 
	    write_name_file("collection.names", current_files);

	cmd = "cd " + backup_root + 
	    ";name_file=" + name_file_name + ";" + backup_cmd;

    } else {
	cmd = "cd " + backup_root + ";" + backup_cmd;
    }
    FILE *pipe_file = popen(cmd.c_str(), "r");
    if (pipe_file == NULL) 
	SYSTEM_ERROR("Could not open pipe from " + cmd);

    // The file to write the slice to
    int collect_file = open(batch_file.c_str(), O_WRONLY|O_CREAT, 0600);
    if (collect_file <= 0)
	SYSTEM_ERROR("Could not create " + batch_file);

    // The size of the data copied (adjusted for the stuff we do automatically)
    size_t copy_size = sizeof(struct file_head) + sizeof(struct file_tail);

    const unsigned int BUFFER_LIMIT = 60*1024;
    // The buffer for copying the files
    char *buffer = new char[BUFFER_LIMIT];	//TODO2: Make configurable

    struct file_head head = {0};
    head.magic = HEAD_MAGIC;
    head.struct_size = sizeof(head);
    head.backup_number = backup_serial();
    if (write(collect_file, &head, sizeof(head)) != sizeof(head))
	SYSTEM_ERROR("Could not write slice output file");
    
    while (1) {
	// The size of the actual read
	int size = fread(buffer, 1, BUFFER_LIMIT, pipe_file);
	if (size <= 0) {
	    break;
	}
	if (write(collect_file, buffer, size) != size)
	    SYSTEM_ERROR("Could not write collection output file");
	copy_size += size;
	head.size += size;
	head.crc = crc32(head.crc, buffer, size);
    }
    delete [] buffer;

    struct file_tail tail = {0};
    tail.magic = TAIL_MAGIC;
    tail.struct_size = sizeof(tail);
    if (write(collect_file, &tail, sizeof(tail)) != sizeof(tail))
	SYSTEM_ERROR("Could not write collection output file");

    if (lseek(collect_file, 0, SEEK_SET) != 0L)
	SYSTEM_ERROR("Could not seek on collection output file");

    if (write(collect_file, &head, sizeof(head)) != sizeof(head))
	SYSTEM_ERROR("Could not write colletion output file");

    close(collect_file);
    pclose(pipe_file);

    if (names_needed) {
	if (! keep_option) {
	    unlink(name_file_name.c_str());
	}
    } else {
	done = true;
    }

    // Information about this batch
    collection_batch_info_b *info = 
	new collection_batch_info_b(current_files, batch_file);

    info->backup_cmd = backup_cmd;
    info->restore_cmd = restore_cmd;
    info->verify_cmd = verify_cmd;
    info->files = current_files;
    info->batch_number = next_batch_number;
    info->names_needed = names_needed;
    info->name = name;
    ++next_batch_number;

    current_size = 0;
    current_files.clear();
    return (info);
}

/********************************************************
 * collection_batch_info::write -- 			*
 * 	Write out the information to a file.		*
 ********************************************************/
void collection_batch_info_b::write(
   std::ostream& out_file	// File to write the stuff to
) const {
    write_base(out_file);
    out_file << META_EOR << std::endl;
}
