/********************************************************
 * config -- The configuration information classes	*
 * 							*
 * Classes:						*
 * 	single_config -- A single configuration 	*
 * 		information set.			*
 * 	config_set -- All the configurations		*
 * 		for all the world
 ********************************************************/
#ifndef __CONFIG_H__
#define __CONFIG_H__
#include <string>
#include <iostream>
#include <vector>
#include <map>

#include "exception.h"
#include "types.h"

/********************************************************
 * VALID_FLAGS -- Flags used for validation		*
 ********************************************************/
const int VALID_FLAG_MANY =     (1<<0);	// Allow multiple entries
					// (Else only single allowed)
const int VALID_FLAG_NUMERIC =  (1<<1);	// Numeric values only
					// (Not used yet)
const int VALID_FLAG_REQUIRED = (1<<2);	// Required parameter

const int VALID_FLAG_SIZE =     (1<<3);	// Size type parameter
					// (Not used yet)

/********************************************************
 * valid_table -- Table of items to validate a 		*
 * 	configuration.					*
 ********************************************************/
struct valid_table {
    const char* keyword;	// Name of the configuration item
    unsigned int flags;		// Validation flags
};


class config;
/********************************************************
 * single_config -- A single entry in the configuration	*
 * table.						*
 ********************************************************/
class single_config {
    private:
	/* 
	 * Stl helper class to do a case insensitive key
	 * comparison between two string
	 */
	class str_less {
	    public:
		bool operator() (
		   const std::string& a,
		   const std::string& b
		) const {
		    return (strcasecmp(a.c_str(), b.c_str()) < 0);
		}
	};
	// Given a configuration keyword, return
	// all the strings that match it.
	typedef std::map<std::string, string_vector, str_less> ConfigInfo;

	std::string name;	// Name of the configuration information
	std::string type; 	// Type of the configuration information

	ConfigInfo data;	// Configuration data
    public:
	// Constructor defaults
	// Copy constructor defaults
	// Assignment operator defaults
	// Destructor defaults
    public:
	void read(
		std::istream &in_file,   // Stream to read
		const std::string i_type,// Name of the configuration
		const std::string i_name // Type of configuration
	);

	void validate(void);	// Validate the data

	// Return the value of a key
	const std::string &value(
		const std::string key,	// Keyword
		unsigned int index = 0	// Index into the key
	) {
	    if (!defined(key, index)) {
		USER_ERROR(
			"Attempt to use undefined key: " + key +
			" (type = " + type + " name= " + name + ")" );
	    }
	    return (data[key][index]);
	}

	// Return the all the values for a key
	const string_vector& get_list(
		const std::string key	// Keyword
	) {
	    if (!defined(key, 0)) {
		USER_ERROR(
			"Attempt to use undefined key: " + key +
			" (type = " + type + " name= " + name + ")" );
	    }
	    return (data[key]);
	}

	// Return true if an option is defined
	const bool defined(
		const std::string key,		// Keyword for the option
		const unsigned int index = 0	// The index for the option
	) const {
	    // The location of the vector containing the option
	    ConfigInfo::const_iterator option_vector = data.find(key);

	    if (option_vector == data.end()) 
		return (false);

	    return (option_vector->second.size() > index);
	}

	void print(std::ostream &out_file) const;

	// Get option -- Single value
	const std::string& get_option(
		const std::string& keyword,
		const unsigned int index = 0
	) const;

	// Get an option -- all values
	const string_vector get_option_all(
		const std::string& keyword
	) const;

    friend class config;
};

/********************************************************
 * config -- A configuration table 			*
 ********************************************************/
// Hierarchy of configs
// 	Type -> Name -> Simple
class config {
    private:
	// Map  Name->SimpleConfig
	typedef std::map<std::string, single_config> config_set;

	// Map  Type->config_set (Type->Name->Simple)
	typedef std::map<std::string, config_set> master_config_type;

	// The configuration data
	master_config_type config;
    public:
	// Constructor defaults
	// Copy Constructor defaults
	// Assignment operator defaults
	// Destructor Defaults
	void read(const char* const file_name);
	void print(std::ostream &out_file) const;

	// Get an option (throw exception if not there)
	const std::string &get_option(
		const std::string& type,	// Type name for lookup
		const std::string& name,	// Name of the option for lookup
		const std::string& option,	// Option name
		const unsigned int index = 0	// Index of the options
	    ) const;

	// Test to see if an option is defined
	const bool defined(
		const std::string& type,	// Type name for lookup
		const std::string& name,	// Name of the option for lookup
		const std::string& option,	// Option name
		const unsigned int index = 0	// Index of the options
	    ) const;

	const single_config &get_single(
		const std::string& type,	// Type name for lookup
		const std::string& name		// Name of the option for lookup
	    ) const;

	void validate(
	    const std::string& type,	// Type name 
	    const std::string& name,	// Option name
	    const valid_table table[]	// Validation table
	);

};

#endif /* __CONFIG_H__ */
