/********************************************************
 * proc -- Module to handle the various sub-processes	*
 * we fork off.  Will record I/O and status for all 	*
 * these processes.  					*
 * 							*
 * This module is designed to keep multiple background	*
 * processes running at once.				*
 ********************************************************/
#ifndef __PROC_H__
#define __PROC_H__

#include <string>
#include <unistd.h>

// Generic callback function (internal to proc module)
typedef void cb_funct(void* const);

// Function to call when a command terminates
typedef void sys_cb_funct(
    pid_t pid, 	// Child process number
    int status	// Status of the child
);


/********************************************************
 * do_system -- Execute a system command and wait 	*
 * for the results.					*
 ********************************************************/
extern int do_system(
   const std::string cmd,	// Command to execute
   const char log_ch		// Character for logging
);
/********************************************************
 * A Class for reading an I/O stream and recording	*
 * the results in a string.				*
 * 							*
 * Private to the proc module.				*
 ********************************************************/
class io_str {
    private:
	const int fd;		// File descriptor we are watching
	std::string str;	// String containing the data from the I/O
	const char log_char; 	// Character for logging
	std::string pid_msg;	// A string indicating the current pid
    public:
	io_str(
	       pid_t pid,		
	       const int i_fd, 
	       const char i_log_char
        );
	io_str(void): fd(-1) {}
	// Copy constructor defaults
	// Assignment operator defaults
	~io_str();
    public:
	static void do_cb_static(void* const me) {
	    static_cast<io_str* const>(me)->do_cb();
	}
    private:
	void do_cb(void);
};


/********************************************************
 * system_info -- Information about a procedure we 	*
 * are waiting on.					*
 * 							*
 * When the procedure exits, we will call the CB 	*
 * function, and then go away.				*
 ********************************************************/
class system_info {
    private:
	const char log_char;	// Where we log things
	const int out_fd;	// FD for stdout of child process
	const int err_fd;	// FD for stderr of child process
	io_str out_str;		// String handler for stdout 
	io_str err_str;		// String handler for stderr

	// Function to call when the item terminates
        sys_cb_funct* const cb_funct; 
    public:
	const pid_t pid;	// Pid of the system call
	bool done;		// True if the process was done
	int status;		// Status of the process
    public:
	system_info(
	    const pid_t i_pid,	    // Process ID of the program
	    const char i_log_char,  // Where we log things
	    const int i_out_fd,	    // FD for standard out
	    const int i_err_fd,	    // Error file descriptor

	    // Function to call when the item terminates
	    sys_cb_funct* const i_cb_funct
	);
	system_info(
	    const pid_t i_pid,	    // Process ID of the program
	    const char i_log_char,  // Where we log things
	    const int i_out_fd,	    // FD for standard out

	    // Function to call when the item terminates
	    sys_cb_funct* const i_cb_funct
	);
	~system_info() {
	    if (out_fd >= 0)
		close(out_fd);
	    if (err_fd >= 0)
		close(err_fd);
	}
    private:
	// No copy constructors
	system_info(const system_info&);

	// No assignment operator
	system_info &operator = (const system_info&);
    public:
	// Called when a SIGCHLD triggers for our process
	void cb(const pid_t cb_pid, const int cb_status) {
	    status = cb_status;
	    done = true;
	    cb_funct(cb_pid, status);
	}
};
#endif /* __PROC_H__ */
