/********************************************************
 * slice_batch -- Module to handle restoration of	*
 * batches created by slicing.				*
 ********************************************************/
#include <fstream>
#include <unistd.h>
#include <sys/fcntl.h>

#include "log.h"
#include "magic.h"
#include "util.h"
#include "colon.h"
#include "restore.h"
#include "slice_batch_r.h"

/********************************************************
 * slice_batch_info_r -- Create a slice batch info	*
 * and restore the data for it from the input file	*
 ********************************************************/
slice_batch_info_r::slice_batch_info_r(
    const batch_info_r& info,		// Generic information about the batch
    std::istream& in_file		// File containing restore data
): batch_info_r(info)
{
    // Parser for the information on this batch
    colon_line parser(in_file);

    parser.expect_line(META_BATCH_SLICE, "Error reading slice batch info",
	    __FILE__, __LINE__);
    sequence = parser.get_int_value("Error reading slice batch info");

    parser.expect_line(META_BATCH_LAST, "Error reading slice batch info",
	    __FILE__, __LINE__);
    if (parser.get_int_value("Error reading slice batch info"))
	last = true;
    else
	last = false;

    std::string line;	// Line from the input file

    std::getline(in_file, line);
    if (line != META_EOR) 
	INTERNAL_ERROR("Error: Tar batch information missing EOR");
}
/********************************************************
 * verify -- verify that the data on the backup is the	*
 * same as the data on the disk.			*
 ********************************************************/
void slice_batch_r::verify(
    const meta_r::info_set& gen_info,// Information about this batch
    io_r* const io 		     // The I/O to use for the verify
) {
    const slice_batch_info_r* const info = 
	dynamic_cast<const slice_batch_info_r* const>
	    ((gen_info.begin())->batch_info);
    
    std::string cmd; // Command to give to the pipe

    if (override_option) {
	// Command to do the restore
	std::string verify_cmd = global_config.get_option(SECTION_BATCH, 
		info->name, 
		BATCH_CMD_VERIFY_CMD);
	cmd = "cd " + backup_root + ";" + verify_cmd;	
    } else {
	cmd = "cd " + backup_root + ";" + info->verify_cmd;	
    }

    verify_restore(gen_info, cmd, io);
}
/********************************************************
 * verify_restore -- Do a restore or verify		*
 ********************************************************/
void slice_batch_r::verify_restore(
    const meta_r::info_set& info,// Information about the batch
    std::string& cmd,		 // Command to use for restore/verify
    io_r* const io 	         // The I/O to use for the verify
)
{
    /* Loop through the items in the information array */
    for (meta_r::info_set::const_iterator cur_info = info.begin();
	    cur_info != info.end();
	    ++cur_info) {

	const slice_batch_info_r* const slice_info = 
	    dynamic_cast<const slice_batch_info_r* const>(cur_info->batch_info);

	// The name of the file containing the batch
	std::string file_name = io->get_file(cur_info->io_info);

	if (slice_info->sequence == 0) {
	    log_msg(LOG_VERIFY, "Pipe command: " + cmd);

	    pipe_file = popen(cmd.c_str(), "w");
	    if (pipe_file == NULL) 
		SYSTEM_ERROR("Could not create pipe for slice");
	}
	log_msg(LOG_RESTORE, "Sending " + file_name + " to pipe");

	// Open a file for this slice
	int in_fd = open(file_name.c_str(), O_RDONLY);
	if (in_fd <= 0)
	    SYSTEM_ERROR("Could not open slice file");

	struct file_head head;	// Head from the file

	if (read(in_fd, &head, sizeof(head)) != sizeof(head))
	    SYSTEM_ERROR("Error reading file header");

	// Bytes left to read
	long unsigned int bytes_left = head.size;

	//TODO2: Validate slice number / last

	while (bytes_left > 0) {
	    //TODO3: Make configurable the size below
	    char buffer[62*1024];	// The buffer for the data

	    // Size of the data to read
	    unsigned long int to_read;

	    if (bytes_left < sizeof(buffer)) 
		to_read = bytes_left;
	    else 
		to_read = sizeof(buffer);

	    // Size of the last read
	    int read_size = read(in_fd, buffer, to_read);

	    if (read_size <= 0)
		break;

	    if (std::fwrite(buffer, 1, read_size, pipe_file) != 
		    static_cast<unsigned int>(read_size))
		SYSTEM_ERROR("Could not write pipe");
	    bytes_left -= read_size;
	}
	close(in_fd);
	fflush(pipe_file);

	if (slice_info->last) {
	    log_msg(LOG_RESTORE, "Closing pipe");
	    pclose(pipe_file);
	    pipe_file = NULL;
	}
	if (! keep_option)
	    io->rm_file(file_name);
    }
}
/********************************************************
 * restore -- Restore a batch of files			*
 ********************************************************/
void slice_batch_r::restore(
     const meta_r::info_set& set,	// Information on what to restore
     io_r* const io,			// I/O to use for restore
     const string_set& files		// Files to restore
) {
    // The information about the first slice
    const slice_batch_info_r* const info = 
	dynamic_cast<const slice_batch_info_r* const>
	((set.begin())->batch_info);
    
    // Files in this batch
    string_set files_this_batch;

    std::set_intersection(files.begin(), files.end(),
	    info->files.begin(), info->files.end(),
		std::insert_iterator<string_set>(
			files_this_batch, files_this_batch.begin()));

    assert(files_this_batch.size() != 0);

    // The name of the file containing the files
    name_file = write_name_file("rest.names", files_this_batch);

    std::string cmd; // Command to do the restore

    if (override_option) {
	std::string restore_cmd = global_config.get_option(SECTION_BATCH, 
		info->name, 
		BATCH_CMD_RESTORE_CMD);
	cmd = "name_file=" + name_file + ";" + restore_cmd;
    } else {
	cmd = "name_file=" + name_file + ";" + info->restore_cmd;
    }

    verify_restore(set, cmd, io);
    if (! keep_option) {
	if (unlink(name_file.c_str()) != 0)
	    SYSTEM_ERROR("Could not unlink " + name_file);
    }
}
