/********************************************************
 * tape_io_b -- Stuff to do a tape I/O backup		*
 ********************************************************/
#include <stdint.h>
#include <cstdlib>
#include <limits.h>
#include <unistd.h>
#include <sys/fcntl.h>

#include "log.h"
#include "proc.h"
#include "magic.h"
#include "util.h"
#include "backup.h"
#include "tmp_file.h"
#include "raw_tape.h"
#include "tty_lock.h"
#include "tape_file.h"
#include "tape_null.h"
#include "tape_tape.h"
#include "tape_io_c.h"
#include "tape_io_b.h"

int tape_io_b::next_tape_number = 1;	// The next tape number

/********************************************************
 * tape_label::write -- Write a tape label to a tape	*
 ********************************************************/
void tape_label::write(
    raw_tape* const mt,		// Tape device to use for label
    const std::string& device,	// The device we are writing
    const int i_tape_number	// The tape number
)
{
    tape_number = i_tape_number;

    // The name of the label file
    std::string label_file_name = tmp_name("tape.label");

    // The file to put the label in
    std::ofstream label_file(label_file_name.c_str());
    if (label_file.bad()) 
	SYSTEM_ERROR("Could not open label file: " + label_file_name);

    label_file << TAPE_LABEL_MAGIC << std::endl;
    label_file << TAPE_LABEL_SERIAL << ": " << label_backup_serial << std::endl;
    label_file << TAPE_LABEL_DATE << ": " << label_backup_date << std::endl;
    label_file << TAPE_LABEL_TAPE << ": " << tape_number << std::endl;
    label_file.close();

    while (1) {
	if (mt->copy_to_mt(label_file_name, ONE_K) == 0)
	    break;
	
	mt->mt_close(__FILE__, __LINE__);
	tty_lock the_lock;	// Lock the console

	std::cout << "Unable to open tape on " << device << std::endl;
	std::cout << "Insert new tape and press <Enter>:";
	std::string response;
	getline(std::cin, response);
	mt->mt_open(__FILE__, __LINE__);
    }
    if (! keep_option) {
	unlink(label_file_name.c_str());
    }
}

/********************************************************
 * next_tape -- Change the tape				*
 ********************************************************/
//TODO2: Check to make sure change worked and we didn't load
// 	a tape we've already written for this backup
void tape_io_b::next_tape(void)
{
    mt->rewind();
    mt->mt_close(__FILE__, __LINE__);

    if (mt->change_tape(next_tape_number))
	return;

    if (change_script != "") {
	tty_lock the_lock;	// Lock the console

	if (do_system(change_script + " " + device + " " +
		int_to_string(next_tape_number),
		LOG_IO) != 0) 
	    USER_ERROR("Change tape script failed");

	if (mt->mt_open(__FILE__, __LINE__) != 0)
	    SYSTEM_ERROR("Could not open " + device);

	return;
    }
    while (1) {
	try {
	    tty_lock the_lock;	// Lock the console

	    log_msg(LOG_IO, "Starting tape " + int_to_string(next_tape_number));
	    std::cout << "Insert tape number " << next_tape_number << 
		" in " << device << " and press<Enter>:";
	    std::string response;
	    getline(std::cin, response);
	    if (mt->mt_open(__FILE__, __LINE__) == 0)
		break;
	    mt->mt_close(__FILE__, __LINE__);
	}
	catch (...) {
	    continue;
	    mt->mt_close(__FILE__, __LINE__);
	}
    }
}
/********************************************************
 * start_tape -- Start writing on the tape		*
 ********************************************************/
void tape_io_b::start_tape(void)
{
    // Write label and zero things out
    mt->rewind();
    label.write(mt, device, next_tape_number);
    ++next_tape_number;
    file_count = 1;	// We start at 1 because of the tape label
}

/********************************************************
 * write -- Write a batch to the tape.			*
 ********************************************************/
void tape_io_b::write(
    const batch_info_b* const batch_info// Information about the batch
) 
{
    if (first) {
	start_tape();
	first = false;
    }

    // The information about this tape
    tape_io_info_b* io_info = 
	new tape_io_info_b(label.tape_number, file_count);

    fill_in_head(batch_info->batch_file, io_info);
    // Status of the copy
    int status = mt->copy_to_mt(batch_info->batch_file, block_size);

    /*
     * We check the status after we put this batch in our
     * list of batches going to the tape.
     */
    
    save_batches.push_back(meta_b::gen_info(batch_info, io_info));
    if (save_batches.size() > safe_count) {
	// Information on the batch we just wrote
	meta_b::gen_info done_batch = save_batches.front();
	save_batches.erase(save_batches.begin());

	global_meta.add_batch(done_batch.batch_info, done_batch.io_info);
	tape_meta.add_batch(done_batch.batch_info, done_batch.io_info);

	if (! keep_option) 
	    unlink(done_batch.batch_info->batch_file.c_str());
    }

    ++file_count;
    if (status != 0) {
	finish_tape();
    }
}
/********************************************************
 * write_eot -- Write the end of the tape. 		*
 ********************************************************/
void tape_io_b::write_eot(
    const bool continued	// Is this continued on the next tape
) {
    std::string label_file = tmp_name("eot");

    log_msg(LOG_IO, "Processing eot for " + device);
    if (continued)
	global_meta.save(label_file, META_PART);
    else
	global_meta.save(label_file, META_FULL);

    if (mt->copy_to_mt(label_file, ONE_K) != 0)
	SYSTEM_ERROR("Could not write final tape file for " + device);

    mt->weof();

    if (!keep_option)
	unlink(label_file.c_str());
}

/********************************************************
 * finish_tape -- Write out the last part of a tape	*
 ********************************************************/
void tape_io_b::finish_tape(void) {
    // Save cached files for safe eof
    std::vector<meta_b::gen_info> batches_to_do = save_batches;

    // Adjust the number of files on the tape
    file_count -= save_batches.size();

    mt->rewind();
    mt->asf(file_count);
    write_eot(true);

    next_tape();
    start_tape();

    // Clear out any save batched
    save_batches.clear();
    
    // Write the files to start the next ape
    for (std::vector<meta_b::gen_info>::const_iterator 
		    cur_info = batches_to_do.begin();
	    cur_info != batches_to_do.end();
	    ++cur_info)
    {
	write(cur_info->batch_info);
    }
}
/********************************************************
 * finish -- Write out the final information to the	*
 * tape.						*
 ********************************************************/
void tape_io_b::finish(
    const std::string meta_file, 	// The meta data file
    const std::string meta_report	// The report file
)
{

    while (save_batches.size() > 0) {
	// A batch we've done
	meta_b::gen_info done_batch = save_batches.front();
	save_batches.erase(save_batches.begin());

	global_meta.add_batch(done_batch.batch_info, done_batch.io_info);
	tape_meta.add_batch(done_batch.batch_info, done_batch.io_info);

	if (! keep_option) 
		unlink(done_batch.batch_info->batch_file.c_str());
    }
    global_meta.save(meta_file, META_FULL);
    global_meta.report(meta_report);

    // These are extras, but why not?
    mt->copy_to_mt(meta_file, ONE_K);
    mt->copy_to_mt(meta_report, ONE_K);
    write_eot(false);

    while (save_batches.size() > 0) {
	if (! keep_option) {
	    unlink(save_batches.front().batch_info->batch_file.c_str());
	}
	save_batches.erase(save_batches.begin());
    }
}
