/********************************************************
 * Module to handle reading files from a tape		*
 ********************************************************/
#include <fstream>
#include <unistd.h>

#include "magic.h"
#include "colon.h"
#include "tmp_file.h"
#include "raw_tape.h"
#include "backup.h"
#include "tty_lock.h"
#include "proc.h"
#include "tape_io_r.h"

/********************************************************
 * tape_label::read -- Read a tape label		*
 ********************************************************/
void tape_label::read(
    raw_tape* const mt,		// Tape device to use for label
    const std::string& device	// The device we are writing
)
{
    tape_number = -1;

    // The name of the label file
    std::string label_file_name = tmp_name("tape.label");
    if (mt->rewind() != 0)
	SYSTEM_ERROR("Could not rewind " + device);

    if (mt->copy_from_mt(label_file_name, ONE_K) != 0) {
	log_msg(LOG_WARNING, "Warning: Could not read tape label");
	return;
    }

    // File to read the input file
    std::ifstream in_file(label_file_name.c_str());

    std::string line;	// Line from the input

    std::getline(in_file,line);
    if (line != TAPE_LABEL_MAGIC) {
	log_msg(LOG_WARNING, "Warning: Tape label: Bad magic number");
	if (! keep_option) {
	    unlink(label_file_name.c_str());
	}
	return;
    }

    colon_line parser(in_file);

    parser.expect_line(TAPE_LABEL_SERIAL, "Error reading tape serial number",
	    __FILE__, __LINE__);
    label_backup_serial = parser.get_int_value(
	    "Error reading tape serial number");

    parser.expect_line(TAPE_LABEL_DATE, "Error reading tape date",
	    __FILE__, __LINE__);
    label_backup_date = parser.get_value();

    parser.expect_line(TAPE_LABEL_TAPE, 
	    "Error reading tape label -- tape number",
	    __FILE__, __LINE__);
    tape_number = parser.get_int_value(
	    "Error reading tape label -- tape number");

    if (! keep_option) {
	unlink(label_file_name.c_str());
    }
}
/********************************************************
 * get_meta -- Read the meta data off the device	*
 * 	and return the result				*
 ********************************************************/
std::string tape_io_r::get_meta(void)
{
    // Name of the file containing the meta data
    std::string meta_file = tmp_name("tape.meta");

    if (mt->rewind() != 0) {
	SYSTEM_ERROR("Could not rewind tape" + device);
    }
    label.read(mt, device);

    if (mt->eod() != 0) {
	SYSTEM_ERROR("Could not space to eot for tape " + device);
    }
    if (mt->bsfm(3) != 0) {
	SYSTEM_ERROR("Could not backup one file for tape " + device);
    }
    mt->copy_from_mt(meta_file, ONE_K);

    return (meta_file);
}
/********************************************************
 * load_tape -- Load the given tape			*
 ********************************************************/
void tape_io_r::load_tape(
   const int tape_number	// Tape number to load
) {
    while (label.tape_number != tape_number) {
	mt->rewind();
	mt->mt_close(__FILE__, __LINE__);

	label.tape_number = -1;

	if (mt->change_tape(tape_number)) {
	    if (mt->mt_open(__FILE__, __LINE__) != 0)
		SYSTEM_ERROR("Could not open tape");

	    label.read(mt, device);
	    if (label.tape_number != tape_number)
		USER_ERROR("Tape system loaded the wrong tape");
	    return;
	}

	if (change_script != "") {
	    tty_lock the_lock;	// Lock the console

	    if (do_system(change_script + " " + device + " " +
		    int_to_string(tape_number),
		    LOG_IO) != 0) 
		USER_ERROR("Change tape script failed");

	    if (mt->mt_open(__FILE__, __LINE__) != 0)
		SYSTEM_ERROR("Could not open " + device);

	    label.read(mt, device);
	    if (label.tape_number != tape_number)
		USER_ERROR("Script loaded the wrong tape. Expect: " +
			int_to_string(tape_number) + " got " +
			int_to_string(label.tape_number));
	    return;
	}

	tty_lock the_lock;	// Lock the console
	std::cout << "Load tape #" << tape_number << 
	    " and press <enter>:";

	std::string response;
	getline(std::cin, response);
	if (mt->mt_open(__FILE__, __LINE__) != 0)
	    continue;
	label.read(mt, device);
    }
}

/********************************************************
 * get_file -- Get a file from tape and put it on disk	*
 ********************************************************/
std::string tape_io_r::get_file(
    const io_info_r* const info		// The information about the batch
) {
    // The information about the CD
    const tape_io_info_r* const tape_info = 
	dynamic_cast<const tape_io_info_r* const>(info);

    load_tape(tape_info->media_sequence);
    mt->asf(tape_info->file_sequence);	

    std::string tmp_file = tmp_name("mt");
    mt->copy_from_mt(tmp_file, block_size);
    validate_file(tmp_file, info);
    return (tmp_file);
}
/********************************************************
 * rm_file -- Remove a tape file that we had previously	*
 * stuck on disk.					*
 ********************************************************/
void tape_io_r::rm_file(
    const std::string& file_name	// Name of the file to remove
) {
    if (! keep_option) {
	if (unlink(file_name.c_str()) != 0)
	    SYSTEM_ERROR("Unable to remove " + file_name);
    }
}
/********************************************************
 * get_cur_media -- Return the currently loaded media	*
 * 	number.						*
 ********************************************************/
int tape_io_r::get_cur_media(void) 
{
    if (label.tape_number < 0)
	label.read(mt, device);
    return (label.tape_number);
}
/********************************************************
 * read -- Read the meta data from the file for the 	*
 * tape/io record.  The generic information has already	*
 * been read.						*
 ********************************************************/
void tape_io_info_r::read(
    std::istream& in_file		// The file to read
) 
{
    // Actually, it's all taken care of in the generic part
}



