/********************************************************
 * tape_tape -- Handle low level I/O to a  tape.	*
 ********************************************************/
#include <fcntl.h>
#include <sys/mtio.h>
#include <limits.h>

#include "log.h"
#include "util.h"
#include "proc.h"
#include "tape_tape.h"

/********************************************************
 * tape_tape -- Create a new tape I/O class		*
 ********************************************************/
tape_tape::tape_tape(
    const std::string& i_device,
    const int i_mode
): device(i_device), mode(i_mode)
{
    if (getenv("DEBUG_TAPE_LIMIT") != NULL) {
	file_limit = atoi(getenv("DEBUG_TAPE_LIMIT"));
    } else {
	file_limit = INT_MAX;
    }
    tape_fd = -1;
    if (mt_open(__FILE__, __LINE__) != 0)
	SYSTEM_ERROR("Could not open " + device);
    rewind();
}
/********************************************************
 * copy_to_mt -- Copy a file to the tape using		*
 * 	the given block size.				*
 * 							*
 * Returns 0 or the error number if an error occurred	*
 ********************************************************/
int tape_tape::copy_to_mt(
    const std::string& file_name, // Name of the file to go to tape
    const int block_size	  // Block size for copy
)
{
    log_msg(LOG_TAPE, "Copy to tape " + file_name + "->" + device + " Blocking " +
	    int_to_string(block_size) + " File " + int_to_string(file_number));

    // Open a fd for the file we are reading
    int in_fd = open(file_name.c_str(), O_RDONLY);
    if (in_fd < 0) 
	SYSTEM_ERROR("Could not open " + file_name + " for reading");

    // Create a buffer for the data
    char *buffer = new char[block_size];
    while (1) {
	// Clean out the buffer (for padding at end)
	memset(buffer, '\0', block_size);

	// Read the data
	int read_size = read(in_fd, buffer, block_size);
	if (read_size <= 0)
	    break;

	if (write(tape_fd, buffer, block_size) != block_size) {
	    int error_number = errno;
	    log_msg(LOG_FULL, "Could not write to tape errno=" + int_to_string(error_number) + " File = " + int_to_string(file_number));
	    delete[] buffer;
	    file_number = -1;
	    close(in_fd);
	    return error_number;
	}
    }
    delete[] buffer;
    close(in_fd);
    return(weof());
}
/********************************************************
 * copy_from_mt -- Copy a file from the tape using	*
 * 	the given block size.				*
 * 							*
 * Returns 0 or the error number if an error occurred	*
 ********************************************************/
int tape_tape::copy_from_mt(
    const std::string& file_name, // File to put the data in
    const int block_size	  // Block size for copy
)
{
    log_msg(LOG_TAPE, "Copy from tape " + device + "->" + file_name + " blocking " + int_to_string(block_size));

    // Open a fd for the file we are reading
    int out_fd = open(file_name.c_str(), O_CREAT|O_WRONLY, 0700);
    if (out_fd < 0) 
	SYSTEM_ERROR("Could not open " + file_name + " for writing");

    // Create a buffer for the data
    char *buffer = new char[block_size];
    while (1) {
	// Clean out the buffer (for padding at end)
	memset(buffer, '\0', block_size);

	// Read the data
	int read_size = read(tape_fd, buffer, block_size);
	if (read_size <= 0)
	    break;

	if (write(out_fd, buffer, block_size) != block_size) {
	    delete[] buffer;
	    file_number = -1;
	    return errno;
	}
    }
    delete[] buffer;
    if (file_number >= 0)
	++file_number;
    return (0);
}
/********************************************************
 * weof -- Write an eof to the tape			*
 ********************************************************/
int tape_tape::weof(void) 
{
    static struct mtop weof_of = {
	MTWEOF,
	1
    };

    ++file_number;
    if (file_number > file_limit)
	return (EIO);

    log_msg(LOG_TAPE, "Writing eof " + device);
    return (ioctl(tape_fd, MTIOCTOP, &weof_of));
}

/********************************************************
 * rewind -- Rewind the tape				*
 ********************************************************/
int tape_tape::rewind(void) 
{
    static struct mtop rew = {
	MTREW,
	1
    };

    file_number = 0;
    log_msg(LOG_TAPE, "Rewind " + device);
    return (ioctl(tape_fd, MTIOCTOP, &rew));
}

/********************************************************
 * eod -- Space to the end of data			*
 ********************************************************/
int tape_tape::eod(void) 
{
    static struct mtop eod_op = {
	MTEOM,
	1
    };

    file_number = -1;
    log_msg(LOG_TAPE, "EOD " + device);
    return (ioctl(tape_fd, MTIOCTOP, &eod_op));
}
/********************************************************
 * bsfm -- Space back filemarks				*
 ********************************************************/
int tape_tape::bsfm(
    const unsigned int count	// Number of file marks to backspace
) 
{
    static struct mtop bsfm_op = {
	MTBSFM,
	1
    };

    bsfm_op.mt_count = count;
    if (file_number > 0)
	file_number -= count;
    log_msg(LOG_TAPE, "BSFM " + device + " files= " + int_to_string(count));
    return (ioctl(tape_fd, MTIOCTOP, &bsfm_op));
}

/********************************************************
 * asf -- Position to the given tape mark		*
 ********************************************************/
int tape_tape::asf(
    const int position	// Location to position to
)
{
    log_msg(LOG_TAPE, "Position " + device + " to file " + 
	    int_to_string(position) + " Current " + int_to_string(file_number));

    if (file_number <= 0) {
	rewind();
    }
    if (position > file_number) {
	static struct mtop fsf = {
	    MTFSF,
	    0
	};
	fsf.mt_count = position - file_number;
	file_number = position;
	log_msg(LOG_TAPE, "Going forward " + int_to_string(fsf.mt_count) + " files");
	return (ioctl(tape_fd, MTIOCTOP, &fsf));
    }
    if (position < file_number) {
	static struct mtop bsfm_op = {
	    MTBSFM,
	    0
	};
	bsfm_op.mt_count = file_number - position;
	file_number = position;
	log_msg(LOG_TAPE, "Going back " + int_to_string(bsfm_op.mt_count) + " files");
	return (ioctl(tape_fd, MTIOCTOP, &bsfm_op));
    }
    return (0);
}

/********************************************************
 * mt_close -- Close a tape drive.			*
 ********************************************************/
void tape_tape::mt_close(const char*const file, const int line) {
    log_msg(LOG_TAPE, std::string("mt_close(file): ") + file + ":" + 
		int_to_string(line));

    if (tape_fd >= 0) {
	close(tape_fd);
	tape_fd = -1;
	file_number = -1;
    } else {
	log_msg(LOG_TAPE, "Extra close: ");
    }
}
/********************************************************
 * mt_open -- Open the tape drive again.		*
 ********************************************************/
int tape_tape::mt_open(const char*const file, const int line) {
    log_msg(LOG_TAPE, std::string("mt_open(file): ") + file + ":" + 
		int_to_string(line));

    assert(tape_fd < 0);
    tape_fd = open(device.c_str(), mode);
    if (tape_fd < 0) {
	int error_number = errno;
	log_msg(LOG_TAPE, "Open for " + device + 
		" failed: " + int_to_string(error_number));
	return (error_number);
    }
    
    if (fcntl(tape_fd, F_SETFD, FD_CLOEXEC) != 0)
	SYSTEM_ERROR("fcntl(" + device + ") failed.");

    return (rewind());
}
