/********************************************************
 * tape_test -- Read all the data from a tape		*
 * 	and report on what's on it.			*
 * 							*
 * Debug tool.						*
 ********************************************************/
#include <iostream>
#include <fstream>
#include <sys/mtio.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <iomanip>

#include "io_c.h"
#include "crc32.h"

/********************************************************
 * mt_file -- Handle a tape file.			*
 * 							*
 * Reads fixed blocks off the tape, buffers for the	*
 * user.						*
 ********************************************************/
class mt_file {
    private:
	int fd;			// FD of the tape to read
	char buffer[64*1024];	// Buffer to read
	char* cur_ptr;		// Where we are in the buffer
	long int bytes_left;	// Number of bytes left in the buffer
	long int last_read_size;// Size of the latest read
    public:
	int file_number;	// Current file number
    public:
	mt_file(const char* const file) {
	    // Open the tape device
	    fd = open(file, O_RDONLY);
	    if (fd < 0) {
		std::cerr << "Error: Could not open " << file << std::endl;
		exit (8);
	    }
	    bytes_left = 0;
	    cur_ptr = buffer;
	    file_number = 0;
	    static struct mtop rew = {
		MTREW,
		1
	    };

	    if (ioctl(fd, MTIOCTOP, &rew) != 0) {
		std::cerr << "ERROR: Could not rewind tape " << std::endl;
		exit (8);
	    }
	}
	~mt_file() {
	    close(fd);
	}
    private:
	mt_file(const mt_file&);
	mt_file& operator=(const mt_file&);
    public:
	// Read a buffer
	long int read(char* const buffer_ptr, const int size);

	// Print a soft error report
	void print_errors(void) {
	    // The tape status
	    struct mtget mt_stat;
	    if (ioctl(fd, MTIOCGET, &mt_stat) != 0) {
		std::cerr << "ERROR: Could not get tape status" << std::endl;
		exit (8);
	    }
	    std::cerr << "Soft error count " << mt_stat.mt_erreg << std::endl;
	}
	void finish_file() {
	    while (last_read_size > 0) {
		// Size of the buffer read
		last_read_size = ::read(fd, buffer, sizeof(buffer));
	    }
	    // Reset input counters
	    cur_ptr = buffer;
	    bytes_left = last_read_size;

	    ++file_number;
	}
};

/********************************************************
 * read -- Read a buffer from the tape.			*
 * 							*
 * Returns: # bytes read.				*
 ********************************************************/
long int mt_file::read(
    char* const buffer_ptr, 	// Buffer to read
    int size			// Size of what we want
)
{
    long int read_size = 0;	// Size of what was read
    char* cur_out_ptr = buffer_ptr;	// Where we're putting the data

    while (size > 0) {
	// Copy the data from the buffer
	
	long int copy_size = bytes_left;	// # bytes to copy
	if (copy_size > size) {
	    copy_size = size;
	}
	if (copy_size > 0)
	    memcpy(cur_out_ptr, cur_ptr, copy_size);

	// Adjust output counters
	cur_out_ptr += copy_size;
	read_size += copy_size;
	size -= copy_size;

	// Adjust input counters
	cur_ptr += copy_size;
	bytes_left -= copy_size;

	// Need another buffer?
	if ((bytes_left == 0) && (size > 0)) {
	    // Read the new buffer

	    // Size of the buffer read
	    last_read_size = ::read(fd, buffer, sizeof(buffer));

	    // I/O error -- abort
	    if (last_read_size < 0) {
		std::cout << "Read error " << errno << std::endl;
		exit(8);
	    }

	    // Hit eof?
	    if (last_read_size == 0) {
		return (read_size);
	    }
	    // Reset input counters
	    cur_ptr = buffer;
	    bytes_left = last_read_size;
	}
    }
    return (read_size);
}

// Status as returned by file_info
enum file_status {FILE_ERROR, FILE_OK, FILE_EOT};

/********************************************************
 * file_info -- Print info about a single mt file	*
 * 							*
 * Returns:						*
 *  	FILE_ERROR -- Could not read file		*
 *  	FILE_OK -- Read file (CRC may not match, but 	*
 *  			we read it)			*
 *  	FILE_EOT -- Hit eot.				*
 ********************************************************/
enum file_status file_info(
    mt_file& in_file		// Tape file to read
)
{
    std::cout << std::setw(3) << in_file.file_number << ": ";

    struct file_head head;	// The head of the file

    int actual_read_size = in_file.read(reinterpret_cast<char*>(&head), sizeof(head));
    if (actual_read_size == 0) {
	return (FILE_EOT);
    }
    if (actual_read_size != sizeof(head))  {
	std::cout << "ERROR: Could not read heading record " << std::endl;
	return(FILE_ERROR);
    } 
    if (head.magic != HEAD_MAGIC)  {
	std::cout << "Bad head magic number " << std::endl;
	return (FILE_OK);
    }

    // Bytes left to read
    long int bytes_to_go = head.size;

    // Crc we've computed so far
    unsigned long int crc = 0;

    while (bytes_to_go > 0) {
	char buffer[60*1024];	// Buffer for the data
	long int read_size = bytes_to_go;	// Size of the next read

	if (read_size > static_cast<long int>(sizeof(buffer)))
	    read_size = sizeof(buffer);

	actual_read_size = in_file.read(buffer, read_size);
	if (actual_read_size != read_size) {
	    std::cout <<"ERROR: Short read " << std::endl;
	    return(FILE_ERROR);
	}

	crc = crc32(crc, buffer, read_size);
	bytes_to_go -= read_size;
    }

    struct file_tail tail;	// The head of the file

    actual_read_size = in_file.read(reinterpret_cast<char*>(&tail), sizeof(tail));
    if (actual_read_size != sizeof(tail))  {
	std::cerr << "ERROR: Could not read tail " << std::endl;
	return (FILE_ERROR);
    }

    if (tail.magic != TAIL_MAGIC) {
	std::cout << "ERROR: Bad tail magic number " << std::endl;
	return (FILE_ERROR);
    }

    std::cout << head.backup_number;
    std::cout << " (" << head.media_sequence << ":" << 
	head.file_sequence << ") ";

    std::cout << "CRC " << head.crc << 
	((crc == head.crc) ? " Good" : " BAD!") <<
	std::endl;
    return (FILE_OK);
}


int main(int argc, char *argv[])
{
    if (argc != 2) {
	std::cerr << "Usage: " << argv[0] << " <tape>\n";
	exit (8);
    }

    mt_file tape(argv[1]);
    while (1) {
	switch (file_info(tape)) {
	    case FILE_ERROR: 
		tape.print_errors();
		exit (8);
	    case FILE_OK:
		break;
	    case FILE_EOT:
		tape.print_errors();
		exit (0);
	    default:
		std::cerr << "Internal error" << std::endl;
		exit (8);
	}
	tape.finish_file();
    }
    return (0);
}

